from os import open as os_open
from stat import S_IRWXG, S_IRWXO, S_IRWXU


def align_int(num: str, align_bytes: int) -> int:
    align_off = num % align_bytes
    if align_off == 0:
        return num
    return num + (align_bytes - align_off)


def executable_opener(path: str, flags: int) -> int:
    return os_open(path, flags, S_IRWXU | S_IRWXG | S_IRWXO)



from abc import abstractmethod
from struct import pack, unpack

class CStruct():
    endian: str

    VALID_ENDIAN = ["<", ">"]

    def __init__(self, endian: str = ">", **kwargs) -> None:
        if endian not in self.VALID_ENDIAN:
            raise ValueError("Invalid endianness")
        self.endian = endian

        for k, v in kwargs.items():
            self.__setattr__(k, v)

    @abstractmethod
    def _struct_vals(self) -> tuple:
        pass

    @abstractmethod
    def _struct_fmt(self) -> str:
        pass

    def _struct_fmt_full(self) -> str:
        return f"{self.endian}{self._struct_fmt()}"

    def pack(self) -> bytes:
        return pack(self._struct_fmt_full(), *map(lambda k : getattr(self, k), self._struct_vals()))

    def unpack(self, buf: bytes) -> None:
        vals = self._struct_vals()
        for i, val in enumerate(unpack(self._struct_fmt_full(), buf)):
            setattr(self, vals[i], val)



from re import X
from slimfat.structs import CStruct


# Definitions in mach-o/fat.h
class FatHeaderStruct(CStruct):
    magic: bytes
    nfat_arch: int

    @staticmethod
    def packsize() -> int:
        return 4 + 4

    def _struct_vals(self) -> tuple:
        return ("magic", "nfat_arch")

    def _struct_fmt(self) -> str:
        return "II"

class FatArchStructBase(CStruct):
    cputype: int
    cpusubtype: int
    offset: int
    size: int
    align: int

    @staticmethod
    def packsize() -> int:
        return 4 + 4

    def _struct_vals(self) -> tuple:
        return ("cputype", "cpusubtype", "offset", "size", "align")

    def _struct_fmt(self) -> str:
        return "ii"

class FatArchStruct(FatArchStructBase):
    def __init__(self, **kwargs) -> None:
        super().__init__(fmt="III", **kwargs)

    @staticmethod
    def packsize() -> int:
        return FatArchStructBase.packsize() + 4 + 4 + 4

    @staticmethod
    def magic() -> int:
        return 0xcafebabe

    def _struct_fmt(self) -> str:
        return f"{super()._struct_fmt()}III"


# This class does not seem to work, but it is exactly as defined to my knowledge...
# objdump, however, recognizes it, so I assume it just isn't implemented just yet
class FatArch64Struct(FatArchStructBase):
    reserved: int = 0

    @staticmethod
    def packsize() -> int:
        return FatArchStructBase.packsize() + 8 + 8 + 4 + 4

    @staticmethod
    def magic() -> int:
        return 0xcafebabf

    def _struct_fmt(self) -> str:
        return f"{super()._struct_fmt()}QQII"

    def _struct_vals(self) -> tuple:
        return (*super()._struct_vals(), "reserved")



from io import SEEK_SET
from os import stat

from slimfat.structs_fat import FatArchStruct, FatHeaderStruct
from slimfat.structs_mach import MachHeaderBegin
from slimfat.util import align_int, executable_opener

def make_fat(output: str, input: list):
    HeaderCls = FatArchStruct
    align = 15
    endian = ">"

    align_bytes = pow(2, align)
    header_len = HeaderCls.packsize() * len(input)

    file_offsets = {}
    with open(output, "wb", opener=executable_opener) as outf:
        outf.write(FatHeaderStruct(endian=endian, magic=HeaderCls.magic(), nfat_arch=len(input)).pack())
        offset = align_int(header_len + outf.tell(), align_bytes)

        for bin in input:
            fstat = stat(bin)

            buf = None
            with open(bin, "rb") as binf:
                buf = binf.read(MachHeaderBegin.packsize())

            machHdr = MachHeaderBegin()
            machHdr.unpack(buf)

            hdr = HeaderCls(endian=endian, cputype=machHdr.cputype, cpusubtype=machHdr.cpusubtype, offset=offset, size=fstat.st_size, align=align)

            file_offsets[bin] = hdr.offset
            offset = align_int(offset + hdr.size, align_bytes)
            outf.write(hdr.pack())

        for bin in input:
            outf.seek(file_offsets[bin], SEEK_SET)
            with open(bin, "rb") as binf:
                outf.write(binf.read())



from slimfat.structs import CStruct


# Defined in mach-o/loader.h
class MachHeaderBegin(CStruct):
    magic: int
    cputype: int
    cpusubtype: int

    VALID_MAGICS: list = [
        0xfeedface,
        0xfeedfacf,
    ]

    def _struct_fmt(self) -> str:
        return "III"

    def _struct_vals(self) -> tuple:
        return ("magic", "cputype", "cpusubtype")

    def unpack(self, buf: bytes) -> None:
        if buf[0] == 0xfe:
            self.endian = ">"
        elif buf[3] == 0xfe:
            self.endian = "<"

        super().unpack(buf)

        if self.magic not in self.VALID_MAGICS:
            raise ValueError("Invalid magic")

    @staticmethod
    def packsize():
        return 4 + 4 + 4
